/*
 *  SH95UPD v0.0.6
 *
 *  KernelEx
 *  Copyright (C) 2006, Xeno86
 *
 *  Original WIN98UPDT0001
 *  Copyright (C) 2005/2006, (x)
 *
 *  This file is part of KernelEx source code.
 *
 *  KernelEx is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; version 2 of the License.
 *
 *  KernelEx is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/** @file file.cpp
 * Class CFile - implementation.
 */

#include "File.h"

/** Constructor. */
CFile::CFile()
{
    m_BufferLen = 0;
    m_Buffer = 0;
    m_MaxBufferLen = 0;
}

/** Destructor. */
CFile::~CFile()
{
    m_BufferLen = 0;
    if ( m_Buffer != NULL )
    {
        free( m_Buffer );
        m_Buffer = NULL;
    }
}

/** Retrieves file size.
 * @param FilePath Path to file.
 * @return Size of file.
 */
int CFile::GetFileSize( const char* FilePath )
{
    FILE*  File = fopen( FilePath, "rb" );
    int FileSize = -1;
    if ( File )
    {
        fseek( File, 0, SEEK_END );
        FileSize = ftell( File );
        fclose( File );
    }

    return FileSize;
}

/** Loads file into memory.
 * @param FilePath Path to file.
 * @param SizeToAdd Additional number of bytes to reserve for file operations.
 * @return True on success, false otherwise.
 */
bool CFile::LoadFromFile( const char* FilePath, int SizeToAdd )
{
    if ( FilePath == NULL )
    {
        return false;
    }

    bool RetCode = false;
    void*  Buffer = NULL;

    int BufferLen = GetFileSize( FilePath );
    if ( BufferLen )
    {
        Buffer = malloc( BufferLen + SizeToAdd);
        if ( Buffer )
        {
            if ( LoadFromFile( FilePath, Buffer, BufferLen ) )
            {
                RetCode = true;
                m_BufferLen = BufferLen;
                m_Buffer = Buffer;
                m_MaxBufferLen = BufferLen + SizeToAdd;
            }
        }
    }

    if ( RetCode == false )
    {
        if ( Buffer != NULL )
        {
            free( Buffer );
        }
    }

    return RetCode;
}

/** Saves file to disk.
 * @param FilePath Path to file.
 * @return True on success, false otherwise.
 */
bool CFile::SaveToFile( const char* FilePath )
{
    return SaveToFile( FilePath, m_Buffer, m_BufferLen );
}

/** Loads file into memory.
 * @param FilePath Path to file.
 * @param Buffer Destination in memory.
 * @param BufferLen Number of bytes to read.
 * @return True on success, false otherwise.
 */
bool CFile::LoadFromFile( const char* FilePath, void* Buffer, int BufferLen )
{
    FILE*  File;
    bool RetCode = false;

    if ( FilePath == NULL || Buffer == NULL )
    {
        return false;
    }

    File = fopen( FilePath, "rb" );
    if ( File )
    {
        if ( fread( Buffer, BufferLen, 1, File ) == 1 )
        {
            RetCode = true;
        }
        fclose( File );
    }

    return RetCode;
}

/** Saves file to disk.
 * @param FilePath Path to file.
 * @param Buffer Source in memory.
 * @param BufferLen Number of bytes to write.
 * @return True on success, false otherwise.
 */
bool CFile::SaveToFile( const char* FilePath, void* Buffer, int BufferLen )
{
    bool RetCode = false;
    FILE*  File;
    if ( FilePath == NULL || Buffer == NULL )
    {
        return false;
    }

    File = fopen( FilePath, "wb" );
    if ( File )
    {
        if ( fwrite( Buffer, BufferLen, 1, File ) == 1 )
        {
            RetCode = true;
        }
        fclose( File );
    }

    return RetCode;
}

/** Retrieves pointer to file loaded into memory.
 * @return Pointer to file loaded into memory.
 */
void *CFile::GetBuffer()
{
    return m_Buffer;
}

/** Retrieves current buffer length.
 * @return Current size of buffer.
 */
int CFile::GetBufferLen()
{
    return m_BufferLen;
}

/** Retrieves original buffer length.
 * @param FilePath Path to file.
 * @return Buffer length
 */int CFile::GetOriginalLen( const char* FilePath )
{
    return GetFileSize( FilePath );
}

/** Sets new buffer length
 * @param NewSize New buffer length.
 * @return True on success, false otherwise.
 */
bool CFile::SetBufferLen(int NewSize)
{
    if ( NewSize >= m_MaxBufferLen )
    {
    	return false;
	}
	m_BufferLen=NewSize;
	return true;
}

