/*
 *  SH95UPD v0.0.6
 *
 *  KernelEx
 *  Copyright (C) 2006-2007, Xeno86
 *
 *  This file is part of KernelEx source code.
 *
 *  KernelEx is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; version 2 of the License.
 *
 *  KernelEx is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/** @file patch.cpp
 * Class patch_item - implementation.
 */

#include "pefile.h"
#include "patch.h"

/** Constructor.
 * @param binary_code The code used in the patch.
 * @param size sizeof(binary_code)
 * @param abs_ofs Absolute pointers. Table containing offsets to data which needs fixing addresses and relocation.
 * @param abs_api Absolute indirect calls and jumps. Table containing offsets and names of external functions (imported by PE file; needing relocation).
 * @param rel_api Relative calls and jumps. Table containing offsets and names of internal functions (exported by PE file).
 * @param hook_api Hooks. Table containing offsets and names of functions we want to export or replace (if already present in PE file).
 * @param custom Optional function called by Process() before copying the code to PE file. @see Process
 */
patch_item::patch_item(char * binary_code, int size, DWORD abs_ofs[], api_entry abs_api[], api_entry rel_api[], api_entry hook_api[], int (* custom)())
{
	this->code = binary_code;
	this->code_size = size;
	this->abs_ofs = abs_ofs;
	this->abs_api = abs_api;
	this->rel_api = rel_api;
	this->hook_api = hook_api;
	this->custom = custom;
	this->Destination = NULL;
}

/** Destructor. */
patch_item::~patch_item()
{
}

/** Returns size of code in current patch_item.
 * @return Size of code in bytes.
 */
int patch_item::GetPatchSize()
{
	return code_size;
}

/** Calculates the size of relocation table size for current patch_item.
 * @param Rva Code start address for size calculation. First 12 least significant bits are used for calculations.
 * @param prevBlRva Holds lastblockRVA value from previous call to this function. @see apply_patches
 * @param lastblocksize Size of previous IMAGE_BASE_RELOCATION block. Used to determine if we need padding.
 * @return Number of bytes needed to hold relocation data for current patch_item.
 */
int patch_item::GetRelocSize(DWORD Rva, DWORD & prevBlRva, int & lastblocksize)
{
	//skip if no entries
	if (!abs_api && !abs_ofs) return 0;

	DWORD lastblockRVA = Rva & 0xfffff000;
	bool closeblock = false;
	if (lastblockRVA != prevBlRva || lastblocksize == 0) closeblock = true;

	int TotalSize = 0;
	DWORD addr;

	int i = 0;
	int j = 0;
	while ((abs_api && abs_api[i].addr) || (abs_ofs && abs_ofs[j]))
	{
		//select lowest address (MIN()) from abs_api or abs_ofs
		if (abs_api && abs_api[i].addr && abs_ofs && abs_ofs[j])
		{
			if (abs_api[i].addr < abs_ofs[j]) addr = abs_api[i++].addr;
			else addr = abs_ofs[j++];
		}
		else if (abs_api && abs_api[i].addr)
		{
			addr = abs_api[i++].addr;
		}
		else // (abs_ofs && abs_ofs[j])
		{
			addr = abs_ofs[j++];
		}

		if (lastblockRVA != (Rva + addr & 0xfffff000) || closeblock)
		{
			if ((lastblocksize - sizeof(IMAGE_BASE_RELOCATION)) % sizeof(DWORD))
			{
				TotalSize += sizeof(WORD); //align to DWORD boundary
			}
			lastblocksize = sizeof(IMAGE_BASE_RELOCATION);
			TotalSize += sizeof(IMAGE_BASE_RELOCATION);
			lastblockRVA = Rva + addr & 0xfffff000;
			closeblock = false;
		}
		lastblocksize += sizeof(WORD);
		TotalSize += sizeof(WORD);
	}
	prevBlRva = lastblockRVA;
	return TotalSize;
}

/** Associates current patch_item with selected PE file.
 * @param dest PE file.
 * @return 0 on error, 1 if ok.
 */
int patch_item::SetWorkFile(CPEFile * dest)
{
	if (Destination) return 0;
	Destination = dest;
	ImgBase = (DWORD)Destination->GetImageBase();
	return 1;
}

/** Adds current patch_item's exports to PE file's export list.
 * @return 0 on error, 1 if ok.
 */
int patch_item::AddExports()
{
	if (!Destination) return 0;
	for (int i = 0 ; hook_api[i].name[0] ; i++)
	{
		if (!Destination->GetApiExport(hook_api[i].name))
		{
			Destination->AddExport(hook_api[i].name);
		}
	}
	return 1;
}

/** Converts a string containing an ordinal number to int.
 * @param s Input string e.g."Ordinal001".
 * @return Ordinal number or zero on error.
 */
int ordinal(const char s[])
{
	//7 is the length of "Ordinal" (num of chars)
	if (!strnicmp(s, "Ordinal", 7)) return atoi(s + 7);
	return 0;
}

/** Recalculates all addresses (functions, data) for current patch_item, inserts it and hooks it in PE file.
 * @param Rva Address where the patch_item will be located in PE file (used for calculations).
 * @param ptr The code will be copied here.
 * @return 0 on error, 1 if ok.
 */
int patch_item::Process(DWORD Rva, char * ptr)
{
	int result;
	if (!Destination) return 0;

	//patch offsets
	if (abs_ofs) for (int i = 0 ; abs_ofs[i] ; i++)
	{
		*(int *)&code[abs_ofs[i]] += ImgBase + Rva;
		Destination->AddReloc(Rva + abs_ofs[i]);
	}

	//patch imports
	if (abs_api) for (int i = 0 ; abs_api[i].addr ; i++)
	{
		int ord;
		//find a separator
		char *dot = strchr(abs_api[i].name, '.');
		if (dot)
		{
			//separate module name from api name
			char module[API_ENTRY_NAME_LEN];
			int len = dot - abs_api[i].name;
			strncpy(module, abs_api[i].name, len);
			module[len] = '\0';

			//check for ordinal number
			if ((ord = ordinal(dot + 1)) != 0)
			{
				*(int *)&code[abs_api[i].addr] = result = (int)Destination->GetApiImport(module, ord);
			}
			else
			{
				*(int *)&code[abs_api[i].addr] = result = (int)Destination->GetApiImport(module, (dot+1));
			}
		}
		else
		{
			//check for ordinal number
			if ((ord=ordinal(abs_api[i].name))!=0)
			{
				*(int *)&code[abs_api[i].addr] = result = (int)Destination->GetApiImport(NULL, ord);
			}
			else
			{
				*(int *)&code[abs_api[i].addr] = result = (int)Destination->GetApiImport(NULL, abs_api[i].name);
			}
		
		}

		Destination->AddReloc(Rva + abs_api[i].addr);
		if (!result)
		{
			vrprintf(1, "ERROR: Couldn't find import \"%s\"\n", abs_api[i].name);
			return 0;
		}
	}

	//patch exports
	if (rel_api) for (int i = 0 ; rel_api[i].addr ; i++)
	{
		int ord;
		//check for ordinal number
		if ((ord = ordinal(rel_api[i].name)) != 0)
		{
			*(int *)&code[rel_api[i].addr] = (result = (int)Destination->GetApiExport(ord)) - (Rva + rel_api[i].addr + 4 + ImgBase);
		}
		else
		{
			*(int *)&code[rel_api[i].addr] = (result = (int)Destination->GetApiExport(rel_api[i].name)) - (Rva + rel_api[i].addr + 4 + ImgBase);
		}

		if (!result)
		{
			vrprintf(1, "ERROR: Couldn't find export \"%s\"\n", rel_api[i].name);
			return 0;
		}
	}

	//call custom procedure
	if (custom)
	{
		if (!custom())
		{
			vrprintf(1, "ERROR: Custom function failed\n");
			return 0;
		}
	}

	//copy code
	memcpy(ptr, code, code_size);

	//hook apis
	for (int i = 0 ; hook_api[i].name[0] ; i++)
	{
		vrprintf(3, "Patching export %s\n", hook_api[i].name);
		int HookPtr;
		result = (int)Destination->GetApiExport(hook_api[i].name, &HookPtr);
		if (!result)
		{
			vrprintf(1, "ERROR: Couldn't find export \"%s\"\n", hook_api[i].name);
			return 0;
		}
		*(int *)HookPtr = Rva + hook_api[i].addr;
	}

	return 1;
}

//---------------------------------------------------------------------------

/** Constructor.
 * @param new_name Target function name.
 * @param existing_name Existing function name.
 */
patch_item_alias::patch_item_alias(const char * new_name, const char * existing_name) : patch_item(NULL, 0, NULL, NULL, NULL, hook_api, NULL)
{
	memset(hook_api, 0, sizeof(hook_api));
	strcpy(hook_api[0].name, new_name);
	strcpy(existing_export, existing_name);
}

/** Adds a new name for existing function (creates an alias).
 * @param Rva dummy
 * @param ptr dummy
 * @return 0 on error, 1 if ok.
 */
int patch_item_alias::Process(DWORD Rva, char * ptr)
{
	int result;
	if (!Destination) return 0;

	if (hook_api[0].name[0])
	{
		vrprintf(3, "Patching export %s\n", hook_api[0].name);
		int HookPtr;
		Destination->GetApiExport(hook_api[0].name, &HookPtr);

		int ord;
		/* check for ordinal number */
		if ((ord=ordinal(existing_export))!=0) /* Added ability to alias to ordinal */
		{
		*(int *)HookPtr = (result = (int)Destination->GetApiExport(ord)) - ImgBase;
		}
		else
		{
		*(int *)HookPtr = (result = (int)Destination->GetApiExport(existing_export)) - ImgBase;
		}
		if (!result)
		{
			vrprintf(1, "ERROR: Couldn't find export \"%s\"\n", existing_export);
			return 0;
		}
	}

	return 1;
}

//---------------------------------------------------------------------------

/** Processes patches and inserts them into the PE file, updating export and relocation tables if needed.
 * @param PEfile PE file.
 * @param patches Table of patch_items.
 * @return PATCH_FAIL on error, PATCH_OK if ok.
 */
int apply_patches(CPEFile & PEfile, patch_item * patches[])
{
	DWORD TotalCodeSize = 0;
	bool have_relocs = PEfile.GetPEHeader()->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size ? true : false;

	//calculate relocation table size
	int lastblocksize = 0; //we use these two variables since we can't use
	DWORD prevRva = 0;     //static variable in GetRelocSize() (different classes!)
	int RelocsSize = 0;
	for (int i = 0 ; patches[i] ; i++)
	{
		RelocsSize += patches[i]->GetRelocSize(TotalCodeSize, prevRva, lastblocksize);
		TotalCodeSize += patches[i]->GetPatchSize();
		patches[i]->SetWorkFile(&PEfile);
		patches[i]->AddExports();
	}

	//alloc space for relocs
	if (have_relocs && !PEfile.AllocRelocTable(RelocsSize))
	{
		PushError("Failed to allocate relocation table!!", 0);
		return PATCH_FAIL;
	}

	//create new section
	int UpdateRVA;
	char * UpdateOfs = NULL;

	if (TotalCodeSize)
	{
		UpdateOfs = (char *) PEfile.AddSection(UPDT_SEC_NAME, TotalCodeSize, &UpdateRVA, IMAGE_SCN_MEM_EXECUTE | IMAGE_SCN_CNT_CODE);

		//check if section creation/increase was successful
		if ((!UpdateOfs) || (!UpdateRVA))
		{
			PushError("Failed to create " UPDT_SEC_NAME " section!", 0);
			return PATCH_FAIL;
		}
		memset(UpdateOfs, 0, TotalCodeSize);
	}

	//rebuild export table
	if (!PEfile.RebuildExportTable(UPDT_SEC_NAME))
	{
		PushError("Failed to Rebuild export table!!", 0);
		return PATCH_FAIL;
	}

	//RebuildExportTable could change section position so we have to recalculate it
	if (TotalCodeSize) UpdateOfs = PEfile.GetSection(UPDT_SEC_NAME)->PointerToRawData + (char *)PEfile.GetBuffer();

	//the actual patching goes here
	for (int i = 0 ; patches[i] ; i++)
	{
		if (!patches[i]->Process(UpdateRVA, UpdateOfs)) return PATCH_FAIL;
		UpdateOfs += patches[i]->GetPatchSize();
		UpdateRVA += patches[i]->GetPatchSize();
	}

	//update reloc table
	if (have_relocs && !PEfile.UpdateRelocTable())
	{
		PushError("Failed to update relocation table!!", 0);
		return PATCH_FAIL;
	}
	return PATCH_OK;
}
