/*
 *  SH95UPD v0.0.6
 *
 *  KernelEx
 *  Copyright (C) 2006-2007, Xeno86
 *
 *  Original WIN98UPDT0001
 *  Copyright (C) 2005/2006, (x)
 *
 *  This file is part of KernelEx source code.
 *
 *  KernelEx is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; version 2 of the License.
 *
 *  KernelEx is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/** @file pefile.cpp
 * Class CPEFile - implementation.
 */

#include "Pefile.h"
#include "vrprintf.h"
#include <algorithm>
#include <imagehlp.h>

#define ALIGN(X,Y) ( ( X + ( Y - 1 ) ) & ( ~ ( Y - 1 ) ) )

/** Default constructor. */
CPEFile::CPEFile()
{
	m_MZHeader = 0;
	m_PEHeader = 0;
	m_SectionHeaders = 0;
	m_RelocTableCapacity = 0;
}

/** Destructor. */
CPEFile::~CPEFile()
{
	m_NewExports.clear();
	m_NewRelocs.clear();
}

/** Opens PE file for manipulation.
 * @param FilePath Path to PE file.
 * @param SizeToAdd Space in bytes to allocate for manipulations (this != final file size !!).
 * @return True on success, false otherwise.
 */
bool CPEFile::LoadFromFile(const char* FilePath, int SizeToAdd)
{
	if ( CFile::LoadFromFile( FilePath, SizeToAdd ) )
	{
		m_MZHeader = (IMAGE_DOS_HEADER *) GetBuffer();
		int FileSize = GetBufferLen();

		if ( ( m_MZHeader->e_lfarlc == 0x40 && m_MZHeader->e_lfanew < FileSize ) && ( m_MZHeader->e_magic == IMAGE_DOS_SIGNATURE ) )
		{
			m_PEHeader= (IMAGE_NT_HEADERS *) ( (int) m_MZHeader + m_MZHeader->e_lfanew );
			if ( ( m_PEHeader->Signature==IMAGE_NT_SIGNATURE ) && ( m_PEHeader->FileHeader.Machine==IMAGE_FILE_MACHINE_I386 ) && ( m_PEHeader->OptionalHeader.Magic==IMAGE_NT_OPTIONAL_HDR_MAGIC ) )
			{
				m_SectionHeaders = (IMAGE_SECTION_HEADER *) ( (int) m_PEHeader + m_PEHeader->FileHeader.SizeOfOptionalHeader + sizeof(IMAGE_FILE_HEADER) + sizeof(DWORD) );
				return true;
			}
		}

	}
	return false;
}

/** Adds new section.
 * @param SectionName Name of new section. Maximum 8 characters.
 * @param SectionSize Size of new section.
 * @param NewSectionRVA Receives new section's RVA.
 * @param Characteristics Section attributes.
 * @return Pointer to new section or NULL on error.
 */
void* CPEFile::AddSection(const char* SectionName, int SectionSize, int* NewSectionRVA, DWORD Characteristics)
{
	if ( ( SectionName==0 ) || (SectionSize==0) || (NewSectionRVA==0) )
	{
		return 0;
	}

	vrprintf(2, "Adding section %s... ", SectionName);
	IMAGE_SECTION_HEADER *LastSection = ( IMAGE_SECTION_HEADER * ) &m_SectionHeaders [ m_PEHeader->FileHeader.NumberOfSections - 1 ];
	IMAGE_SECTION_HEADER *NewSection = ( IMAGE_SECTION_HEADER * ) &m_SectionHeaders [ m_PEHeader->FileHeader.NumberOfSections ];

	//check if there's enough space for new section header
	if ( ( m_PEHeader->OptionalHeader.SizeOfHeaders - ( (int) NewSection - (int) m_MZHeader ) ) < sizeof(IMAGE_SECTION_HEADER) )
	{
		vrprintf(2, "failed!\n");
		return 0;
	}

	m_PEHeader->FileHeader.NumberOfSections++;

	memset( NewSection, 0, sizeof(IMAGE_SECTION_HEADER) );

	strncpy( (char *) NewSection->Name, SectionName, sizeof(NewSection->Name) );

	NewSection->VirtualAddress = ALIGN( LastSection->VirtualAddress + LastSection->Misc.VirtualSize, m_PEHeader->OptionalHeader.SectionAlignment );
	NewSection->Misc.VirtualSize = SectionSize;

	NewSection->PointerToRawData = ALIGN( LastSection->PointerToRawData + LastSection->SizeOfRawData, m_PEHeader->OptionalHeader.FileAlignment );
	NewSection->SizeOfRawData = ALIGN( SectionSize, m_PEHeader->OptionalHeader.FileAlignment );

	NewSection->Characteristics = IMAGE_SCN_MEM_READ | Characteristics;

	if (Characteristics & IMAGE_SCN_CNT_CODE) m_PEHeader->OptionalHeader.SizeOfCode += ALIGN( SectionSize, m_PEHeader->OptionalHeader.FileAlignment );
	if (Characteristics & IMAGE_SCN_CNT_INITIALIZED_DATA) m_PEHeader->OptionalHeader.SizeOfInitializedData += ALIGN( SectionSize, m_PEHeader->OptionalHeader.FileAlignment );
	m_PEHeader->OptionalHeader.SizeOfImage += ALIGN( SectionSize, m_PEHeader->OptionalHeader.SectionAlignment );

	if (!SetBufferLen( NewSection->PointerToRawData + NewSection->SizeOfRawData )) return 0;

	*NewSectionRVA = NewSection->VirtualAddress;
	vrprintf(2, "done!\n");
	return (void *) ( NewSection->PointerToRawData + (int) GetBuffer() );
}

/** Resizes last section.
 * @param SectionHeader Address of section header.
 * @param AddedSectionSize Bytes to add.
 * @param NewSectionRVA Receives RVA of beginning of allocated space.
 * @return Pointer to beginning of allocated space.
 */
void* CPEFile::IncSection(IMAGE_SECTION_HEADER* SectionHeader, int AddedSectionSize, int* NewSectionRVA)
{
	if ( ( SectionHeader==0 ) || (AddedSectionSize==0) || (NewSectionRVA==0) )
	{
		return 0;
	}

	//display section name
	char SectionName[sizeof(SectionHeader->Name) + 1];
	strncpy(SectionName, (char *)SectionHeader->Name, sizeof(SectionHeader->Name));
	SectionName[sizeof(SectionHeader->Name)] = '\0';
	vrprintf(2, "Incrementing section %s\n", SectionName);

	int SectionSize = SectionHeader->Misc.VirtualSize;

	if (SectionHeader->Characteristics & IMAGE_SCN_CNT_CODE) m_PEHeader->OptionalHeader.SizeOfCode -= ALIGN(SectionSize, m_PEHeader->OptionalHeader.FileAlignment);
	if (SectionHeader->Characteristics & IMAGE_SCN_CNT_INITIALIZED_DATA) m_PEHeader->OptionalHeader.SizeOfInitializedData -= ALIGN(SectionSize, m_PEHeader->OptionalHeader.FileAlignment);
	m_PEHeader->OptionalHeader.SizeOfImage -= ALIGN(SectionSize, m_PEHeader->OptionalHeader.SectionAlignment);

	int Rva = SectionHeader->VirtualAddress + SectionSize;
	int Ofs = SectionHeader->PointerToRawData + SectionSize + (int)GetBuffer();

	SectionSize += AddedSectionSize;

	SectionHeader->Misc.VirtualSize = SectionSize;
	SectionHeader->SizeOfRawData = ALIGN(SectionSize, m_PEHeader->OptionalHeader.FileAlignment);

	if (SectionHeader->Characteristics & IMAGE_SCN_CNT_CODE) m_PEHeader->OptionalHeader.SizeOfCode += ALIGN(SectionSize, m_PEHeader->OptionalHeader.FileAlignment);
	if (SectionHeader->Characteristics & IMAGE_SCN_CNT_INITIALIZED_DATA) m_PEHeader->OptionalHeader.SizeOfInitializedData += ALIGN(SectionSize, m_PEHeader->OptionalHeader.FileAlignment);
	m_PEHeader->OptionalHeader.SizeOfImage += ALIGN(SectionSize, m_PEHeader->OptionalHeader.SectionAlignment);

	if (!SetBufferLen(SectionHeader->PointerToRawData + SectionHeader->SizeOfRawData)) return 0;

	*NewSectionRVA = Rva;
	return (void *)Ofs;
}

/** Changes entry point.
 * @param NewEntryPoint Address of new entry point (RVA).
 */
void CPEFile::SetEntryPoint(int NewEntryPoint)
{
	m_PEHeader->OptionalHeader.AddressOfEntryPoint=NewEntryPoint;
}

/** Retrieves address of entry point.
 * @return Address of entry point (RVA).
 */
void* CPEFile::GetEntryPoint()
{
	return (void *) m_PEHeader->OptionalHeader.AddressOfEntryPoint;
}

/** Retrieves preffered base address.
 * @return Preffered base address.
 */
void* CPEFile::GetImageBase()
{
	return (void *) m_PEHeader->OptionalHeader.ImageBase;
}

/** Retrieves pointer to PE header.
 * @return Pointer to PE header.
 */
IMAGE_NT_HEADERS* CPEFile::GetPEHeader()
{
	return m_PEHeader;
}

/** Retrieves pointer to MS-DOS header.
 * @return Pointer to MS-DOS header.
 */
IMAGE_DOS_HEADER* CPEFile::GetMZHeader()
{
	return m_MZHeader;
}

/** Retrieves pointer to loaded file at given RVA.
 * @param Rva RVA.
 * @return Pointer to loaded file at given RVA.
 */
void* CPEFile::RvaToOffset(DWORD Rva)
{
	IMAGE_SECTION_HEADER *Section = m_SectionHeaders;

	for ( int i=0; i < m_PEHeader->FileHeader.NumberOfSections; i++)
	{
		if ( ( Rva >= Section->VirtualAddress ) && ( Rva < ( Section->VirtualAddress + Section->Misc.VirtualSize ) ) )
			return (void *) ( Rva - Section->VirtualAddress + Section->PointerToRawData + (int) m_MZHeader );
		Section++;
	}

	return 0;
}

/** Retrieves pointer to section header for given section number.
 * @param Section Number of section. (1=first).
 * @return Pointer to section header.
 */
IMAGE_SECTION_HEADER* CPEFile::GetSection(int Section)
{
	if ( ( Section > m_PEHeader->FileHeader.NumberOfSections ) || ( Section < 1) ) return 0;
	return &m_SectionHeaders[ Section - 1 ];
}

/** Retrieves pointer to section header for given section name.
 * @param SectionName Section name.
 * @return Pointer to section header.
 */
IMAGE_SECTION_HEADER* CPEFile::GetSection(const char* SectionName)
{
	for ( int i=0; i < m_PEHeader->FileHeader.NumberOfSections; i++ )
	{
		if ( !strnicmp( (const char *) &m_SectionHeaders[i].Name, SectionName, sizeof( m_SectionHeaders[i].Name ) ) )
		{
			return &m_SectionHeaders[ i ];
		}

	}
	return 0;
}

/** Retrieves VA of specified function imported by name in import address table.
 * @param Library DLL name (with or without extension, NULL=parse all imported libraries).
 * @param Function Function name.
 * @return VA of specified function in import address table.
 */
void* CPEFile::GetApiImport(const char* Library, const char* Function)
{
	IMAGE_IMPORT_DESCRIPTOR *Imports = (IMAGE_IMPORT_DESCRIPTOR *) RvaToOffset(m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress);
	while (Imports->Name)
	{
		if (!Library || !strnicmp((char *)RvaToOffset(Imports->Name), Library,strlen(Library)))
		{ 
			DWORD* Names ;
			if (Imports->OriginalFirstThunk) {
			Names = (DWORD *) RvaToOffset( Imports->OriginalFirstThunk );
			} else {
			Names = (DWORD *) RvaToOffset( Imports->FirstThunk );
			}

			DWORD* Pointers = (DWORD *) (Imports->FirstThunk + (int) GetImageBase());
			while (*Names )
			{
				if (!(*Names & 0x80000000)) //skip ordinals
				{
					char *ApiName = (char *) RvaToOffset(*Names) + sizeof(WORD);
					if (!strncmp(ApiName,Function,strlen(Function))) return Pointers;
				}
				Names++;
				Pointers++;
			}

		}
		Imports++;
	}
	return 0;
}

/** Retrieves VA of specified function imported by ordinal number in import address table.
 * @param Library DLL name (with or without extension, NULL=parse all imported libraries).
 * @param Ordinal Ordinal number.
 * @return VA of specified function in import address table.
 */
void* CPEFile::GetApiImport(const char* Library,WORD Ordinal)
{
	IMAGE_IMPORT_DESCRIPTOR *Imports = ( IMAGE_IMPORT_DESCRIPTOR *) RvaToOffset ( m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress );
	WORD debugtemp;

	while (Imports->Name)
	{
		printf("%s",RvaToOffset(Imports->Name));
		if ( ! Library || ! strnicmp( (char *)RvaToOffset(Imports->Name), Library, strlen (Library)))
		{
			DWORD *Names ;
			if ( Imports->OriginalFirstThunk )
			{
				Names = (DWORD *) RvaToOffset( Imports->OriginalFirstThunk );
			}else{
				Names = (DWORD *) RvaToOffset( Imports->FirstThunk );
			}

			DWORD *Pointers = (DWORD *) ( Imports->FirstThunk + (int)GetImageBase() );

			while ( *Names )
			{
				debugtemp=(*Names&0x0000ffff);
				printf("%d\n",debugtemp);
				if ((*Names & 0x80000000) && ((*Names & 0x0000ffff) == Ordinal)) return Pointers;
				Names++;
				Pointers++;
			}

		}
		Imports++;
	}
	return 0;
}

/** Retrieves VA of specified function exported by name.
 * @param Function Function name.
 * @param ApiPtr Receives pointer to function's entry in export address
 *        table (used for edition, default is NULL).
 * @return VA of specified function.
 */
void* CPEFile::GetApiExport(const char* Function, int* ApiPtr)
{
	IMAGE_EXPORT_DIRECTORY *Exports = ( IMAGE_EXPORT_DIRECTORY * ) RvaToOffset( m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress );

	if ( Exports )
	{
		DWORD *Names = (DWORD *) RvaToOffset( Exports->AddressOfNames );

		for ( DWORD i = 0; i < Exports->NumberOfNames; i++ )
		{
			if ( ! strncmp( (char *) RvaToOffset( *Names), Function, strlen(Function) ) )
			{
				WORD *OrdinalTable = (WORD *) RvaToOffset( (Exports->AddressOfNameOrdinals) );
				WORD Index = OrdinalTable[ i ];
				DWORD *FunctionTable = (DWORD *) RvaToOffset( Exports->AddressOfFunctions );
				DWORD Address = (DWORD) FunctionTable[ Index ] + (int) GetImageBase();
				if ( ApiPtr ) *ApiPtr = (int) &FunctionTable[ Index ];
				return (void *) Address;
			}

			Names++;
		}

	}

	return 0;
}

/** Retrieves VA of specified function exported by ordinal.
 * @param Ordinal Ordinal number.
 * @return VA of specified function.
 */
void* CPEFile::GetApiExport(int Ordinal)
{
	IMAGE_EXPORT_DIRECTORY *Exports = ( IMAGE_EXPORT_DIRECTORY * ) RvaToOffset( m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress );

	if ( ( Ordinal > (int) Exports->NumberOfFunctions ) || ( Ordinal < 1 ) )
	{
		return 0;
	}

	DWORD *FunctionTable = (DWORD *) RvaToOffset( Exports->AddressOfFunctions );
	DWORD Address = (DWORD) FunctionTable[ Ordinal - Exports->Base ] + (int)GetImageBase();
	return (void *) Address;
}

/** Changes preffered base address.
 * @param NewBase New preffered base address.
 * @return True if performed, false if relocation data is unavailable.
 */
bool CPEFile::ChangeImageBase(int NewBase)
{

	if ( m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress )
	{
		IMAGE_BASE_RELOCATION *Relocs = ( IMAGE_BASE_RELOCATION * ) RvaToOffset( m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress );

		while ( Relocs->VirtualAddress )
		{

			int Size = Relocs->SizeOfBlock;
			int Base = (int) RvaToOffset( Relocs->VirtualAddress );
			WORD *RelocList = (WORD *) ( (DWORD)Relocs +  sizeof(IMAGE_BASE_RELOCATION) );

			for ( int i = ( ( Size - sizeof(IMAGE_BASE_RELOCATION) ) / 2); i; i--)
			{
				WORD Rel = *RelocList;

				if ( (Rel>>12)==IMAGE_REL_BASED_HIGHLOW )
				{
					DWORD *Ptr = (DWORD *) (Base + (int) (Rel & 0x0fff) );
					*Ptr += NewBase - (int) GetImageBase();
				}

				RelocList++;
			}

			Relocs = (IMAGE_BASE_RELOCATION *) ( (int) Relocs + Size);
		}

		m_PEHeader->OptionalHeader.ImageBase = NewBase;
		return true;
	}
	return false;
}

/** Calculates file checksum.
 * @return File checksum.
 */
DWORD CPEFile::CalculateChecksum()
{
	unsigned long hdrsum;
	DWORD Checksum;
	if (!CheckSumMappedFile(GetMZHeader(), GetBufferLen(), &hdrsum, &Checksum))
	{
		Checksum = 0;
	}
	return Checksum;
}

/** Calculates and writes file checksum. */
void CPEFile::FixChecksum()
{
	IMAGE_NT_HEADERS *pehdr = GetPEHeader();

	if (pehdr->OptionalHeader.CheckSum)
	{
		DWORD Checksum = CalculateChecksum();
		vrprintf(3, "Checksum set to : 0x%08x\n", Checksum);
		pehdr->OptionalHeader.CheckSum = Checksum;
	}
}

/** Adds export name to a stack for later rebuilding of export table.
 * @param ExportName Export name.
 */
void CPEFile::AddExport(const char* ExportName)
{
	m_NewExports.push_back(ExportName);
	vrprintf(3, "Adding export: %s\n", ExportName);
}

/** Rebuilds export table by either rebuilding existing table or building
 *  new export table if there's not enough space in existing one.
 * @param mtd2secname Section name for method 2.
 * @param Flags: E_ADDORINC or E_MUSTCREATE.
 * @return True on success, false otherwise.
 */
bool CPEFile::RebuildExportTable(const char* mtd2secname, int Flags)
{
	// exit if nothing to be done
	if (m_NewExports.empty()) return true;
	// try method one...
	bool ret = RebuildExportTable_MTD1();
	// ... if it fails try the second one
	if (!ret) ret = RebuildExportTable_MTD2( mtd2secname, Flags );
	return ret;
}

/** Calculates size of export table including export names on the stack.
 * @return Number of bytes required to store the export table.
 */
int CPEFile::GetExportTableSize()
{
	//calcula tamanho da export velha
	IMAGE_EXPORT_DIRECTORY *Exports = ( IMAGE_EXPORT_DIRECTORY * ) RvaToOffset( m_PEHeader->OptionalHeader.DataDirectory[ IMAGE_DIRECTORY_ENTRY_EXPORT ].VirtualAddress );
	int SectionSize = sizeof( IMAGE_EXPORT_DIRECTORY );
	SectionSize += strlen( (char *) RvaToOffset( Exports->Name ) ) + 1;
	SectionSize += ( sizeof( DWORD ) + sizeof( WORD ) ) * Exports->NumberOfNames; // functions table + ordinals table
	SectionSize += sizeof( DWORD ) * Exports->NumberOfFunctions;
	int *ApiNames = (int *) RvaToOffset( Exports->AddressOfNames );
	for ( int i = Exports->NumberOfNames; i; --i )
	{
		SectionSize += strlen( (char *) RvaToOffset( *ApiNames++ ) ) + 1;
	}
	//adiciona tamanho de nossas exports
	for ( DWORD i = 0; i < m_NewExports.size(); ++i )
	{
		std::string j = m_NewExports[ i ];
		SectionSize += j.size() + 1 + ( sizeof( DWORD ) * 2 ) + sizeof( WORD );
	}

	return SectionSize;
}

/** Creates a new export table in a new section leaving the old one
 *  untouched. This causes the file to be larger as there are two coppies
 *  of the table in the file (one unused) and changes sections order.
 * @param SectionName Name of export table's section.
 * @param Flags: E_ADDORINC or E_MUSTCREATE.
 * @return True on success, false otherwise.
 */
bool CPEFile::RebuildExportTable_MTD2(const char* SectionName, int Flags)
{
	//exit if there's nothing to do
	if ( m_NewExports.empty() ) return false;

	vrprintf(2, "Attempting to rebuild export table (Method 2)...\n");

	int SectionSize = GetExportTableSize();

	//adiciona ou aumenta seco
	IMAGE_SECTION_HEADER *ExportSection = this->GetSection( SectionName );
	int Rva;
	void *Raw;
	if ( !ExportSection )
	{
		Raw  = AddSection( SectionName, SectionSize, &Rva, IMAGE_SCN_CNT_INITIALIZED_DATA );
	}
	else
	{
		if ( Flags == E_MUSTCREATE ) return false;
		Raw = IncSection( ExportSection, SectionSize, &Rva );
	}

	if (!Raw || !Rva)
	{
		vrprintf(2, "... failed!\n");
		vrprintf(2, "Couldn't create nor update section %s\n", SectionName);
		return false;
	}

	bool ret = BuildExportTable( Raw, Rva );
	if (ret) vrprintf(2, "... done!\n"); else vrprintf(2, "... failed!\n");
	return ret;
}

/** Builds new export table containing old and new entries.
 * @param Raw Destination.
 * @param Rva Destination RVA.
 * @return True on success, false otherwise.
 */
bool CPEFile::BuildExportTable(void* Raw, int Rva)
{
	//exit if there's nothing to do
	if ( m_NewExports.empty() ) return false;

	char *buf = (char *) Raw;

	//get location of old export table
	IMAGE_EXPORT_DIRECTORY *Exports = ( IMAGE_EXPORT_DIRECTORY * ) RvaToOffset( m_PEHeader->OptionalHeader.DataDirectory[ IMAGE_DIRECTORY_ENTRY_EXPORT ].VirtualAddress );
	m_PEHeader->OptionalHeader.DataDirectory[ IMAGE_DIRECTORY_ENTRY_EXPORT ].VirtualAddress = Rva;

	memcpy( buf, Exports, sizeof( IMAGE_EXPORT_DIRECTORY ) );
	Exports = ( IMAGE_EXPORT_DIRECTORY * ) buf;
	buf += sizeof( IMAGE_EXPORT_DIRECTORY );
	Rva += sizeof( IMAGE_EXPORT_DIRECTORY );

	memcpy( buf, RvaToOffset( Exports->AddressOfFunctions ), sizeof( DWORD ) * Exports->NumberOfFunctions );
	Exports->AddressOfFunctions = Rva;
	buf += sizeof( DWORD ) * ( Exports->NumberOfFunctions + m_NewExports.size() );
	Rva += sizeof( DWORD ) * ( Exports->NumberOfFunctions + m_NewExports.size() );

	DWORD *Names = (DWORD *) buf;
	DWORD *OldNames = (DWORD *) RvaToOffset( Exports->AddressOfNames );
	Exports->AddressOfNames = Rva;
	buf += sizeof( DWORD ) * ( Exports->NumberOfNames + m_NewExports.size() );
	Rva += sizeof( DWORD ) * ( Exports->NumberOfNames + m_NewExports.size() );

	WORD *Ordinals = (WORD *) buf;
	WORD *OldOrdinals = (WORD *) RvaToOffset( Exports->AddressOfNameOrdinals );
	Exports->AddressOfNameOrdinals = Rva;
	buf += sizeof( WORD ) * ( Exports->NumberOfNames + m_NewExports.size() );
	Rva += sizeof( WORD ) * ( Exports->NumberOfNames + m_NewExports.size() );

	strcpy( buf, (char *) RvaToOffset( Exports->Name ) );
	Exports->Name = Rva;
	Rva += strlen( buf ) + 1;
	buf += strlen( buf ) + 1;

	//sort the table containing new exports
	sort(m_NewExports.begin(), m_NewExports.end());

	for ( int i = Exports->NumberOfNames; i; --i )
	{
		bool k = false;
		DWORD j;
		for ( j = 0; j < m_NewExports.size(); j++ )
		{
			if( strcmp( (char *) RvaToOffset( *OldNames ), m_NewExports[ j ].c_str() ) > 0 )
			{
				k = true;
				break;
			}
		}
		if ( k )
		{
			*Ordinals++ = Exports->NumberOfFunctions++;
			strcpy( buf, m_NewExports[ j ].c_str() );
			m_NewExports.erase( m_NewExports.begin() + j );
			Exports->NumberOfNames++; i++;
		} else {
			*Ordinals++ = *OldOrdinals++;
			strcpy( buf, (char *) RvaToOffset( *OldNames++ ) );
		}
		*Names++ = Rva;
		Rva += strlen( buf ) + 1;
		buf += strlen( buf ) + 1;
	}

	//update data directory entry
	m_PEHeader->OptionalHeader.DataDirectory[ IMAGE_DIRECTORY_ENTRY_EXPORT ].Size = (int) buf - (int) Raw;

	return true;
}

/** Retrieves pointer to section header associated with given RVA.
 * @param Rva RVA.
 * @return Pointer to section header.
 */
IMAGE_SECTION_HEADER* CPEFile::RvaToSection(DWORD Rva)
{
	IMAGE_SECTION_HEADER* Section = m_SectionHeaders;

	for (int i = 0 ; i < m_PEHeader->FileHeader.NumberOfSections ; i++)
	{
		if ((Rva >= Section->VirtualAddress) && (Rva < (Section->VirtualAddress + Section->Misc.VirtualSize)))
			return Section;
		Section++;
	}
	return NULL;
}

/** Builds new export table in place of old export table, doesn't change
 *  internal file structure. The weak point is that there might not be
 *  enoungh space for additional entries and then other method must be used.
 * @return True on success, false otherwise.
 */
bool CPEFile::RebuildExportTable_MTD1()
{
	//exit if there's nothing to do
	if (m_NewExports.empty()) return false;

	vrprintf(2, "Attempting to rebuild export table (Method 1)...\n");

	int NewExportTableSize = GetExportTableSize();

	DWORD ExportsRVA = m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress;
	IMAGE_SECTION_HEADER* Section = RvaToSection(ExportsRVA);

	// check if there are no other tables in current section AFTER export table
	// if there are - abort [fixme?]
	for (unsigned long i = 0 ; i < m_PEHeader->OptionalHeader.NumberOfRvaAndSizes ; i++)
	{
		if (m_PEHeader->OptionalHeader.DataDirectory[i].VirtualAddress > ExportsRVA && RvaToSection(m_PEHeader->OptionalHeader.DataDirectory[i].VirtualAddress) == Section)
		{
			vrprintf(2, "... failed!\nDetected unsupported table layout!\n");
			return false;
		}
	}

	// check if there is enough room for new export functions
	int MaxSectionSize = (Section + 1)->VirtualAddress - Section->VirtualAddress;
	int MaxTableSize = MaxSectionSize - (ExportsRVA - Section->VirtualAddress);

	// show how much space is needed for new entries
	vrprintf(2, "Space available: %d Bytes, ", MaxTableSize - m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].Size);
	vrprintf(2, "Space needed: %d Bytes\n", NewExportTableSize - m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].Size);
	if (MaxTableSize < NewExportTableSize)
	{
		vrprintf(2, "... failed!\nThere's not enough space for additional entries in current export table!\n");
		return false;
	}

	// Make space for new export table
	int Offset = ALIGN(ExportsRVA - Section->VirtualAddress + NewExportTableSize, m_PEHeader->OptionalHeader.FileAlignment) - Section->SizeOfRawData;
	if (Offset)
	{
		vrprintf(3, "Applying offset of %d\n", Offset);
		char* tmp0 = (char *) malloc(GetBufferLen() - (Section + 1)->PointerToRawData);
		memcpy(tmp0, RvaToOffset((Section + 1)->VirtualAddress), GetBufferLen() - (Section + 1)->PointerToRawData);
		memcpy(RvaToOffset((Section + 1)->VirtualAddress + Offset), tmp0, GetBufferLen() - (Section + 1)->PointerToRawData);
		free(tmp0);
		memset(RvaToOffset((Section + 1)->VirtualAddress), 0, Offset);

		// update offsets
		for (int i = 0 ; i < m_PEHeader->FileHeader.NumberOfSections ; i++)
		{
			if (&m_SectionHeaders[i] <= Section) continue;
			m_SectionHeaders[i].PointerToRawData += Offset;
		}
		Section->SizeOfRawData += Offset;
		if (!SetBufferLen(GetBufferLen() + Offset)) return false;
	}

	// build a new export table
	void * new_table = malloc(NewExportTableSize);
	BuildExportTable(new_table, ExportsRVA);
	memcpy(RvaToOffset(ExportsRVA), new_table, NewExportTableSize);
	free(new_table);

	Section->Misc.VirtualSize = ExportsRVA + NewExportTableSize - Section->VirtualAddress;
	vrprintf(2, "... done!\n");
	return true;
}

/** Adds reloc to a stack for later rebuilding of relocation table.
 * @param Rva RVA.
 */
void CPEFile::AddReloc(DWORD Rva)
{
	m_NewRelocs.push_back(Rva);
	vrprintf(3, "Adding reloc: RVA 0x%08x\n", Rva);
}

/** Preallocates space for relocation table.
 *  Execute this after adding new relocs, but before adding new section.
 * @param BytesToAdd Number of bytes to allocate.
 * @return True on success, false otherwise.
 */
// We know that relocation table (if exists) is at the end of the PE file
// or just before certificate data. So we simply move certificate, update
// sizes and we're done.
bool CPEFile::AllocRelocTable(int BytesToAdd)
{
	vrprintf(2, "Attempting to allocate space in relocation table... ");

	DWORD RelocsRVA = m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress;

	//check if table exists
	if (!RelocsRVA)
	{
		vrprintf(2, "failed! Table not found!\n");
		return false;
	}

	//check of there is debug data inside
	if (m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].VirtualAddress)
	{
		vrprintf(2, "\nError: Debug data is not supported!\n");
		return false;
	}

	IMAGE_SECTION_HEADER* Section = RvaToSection(RelocsRVA);

	int Offset = ALIGN(RelocsRVA - Section->VirtualAddress + m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size + BytesToAdd, m_PEHeader->OptionalHeader.FileAlignment) - Section->SizeOfRawData;
	if (Offset)
	{
		vrprintf(3, "\nApplying offset of %d\n... ", Offset);
		//move certificate
		DWORD ce_PTR = m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_SECURITY].VirtualAddress;
		if (ce_PTR && ce_PTR < Section->PointerToRawData + Section->SizeOfRawData + Offset)
		{
			int ce_size = m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_SECURITY].Size;
			char* tmp0 = (char *) malloc(ce_size);
			memcpy(tmp0, (void *)((DWORD)m_MZHeader + ce_PTR), ce_size);
			memcpy((void *)((DWORD)m_MZHeader + ce_PTR + Offset), tmp0, ce_size);
			free(tmp0);
			//fill with zeros & correct offset
			memset((void *)((DWORD)m_MZHeader + ce_PTR), 0, Offset);
			ce_PTR = m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_SECURITY].VirtualAddress += Offset;
			if (!SetBufferLen(ce_PTR + ce_size)) return false;
		}

		//correct sizes
		m_PEHeader->OptionalHeader.SizeOfImage -= ALIGN(Section->SizeOfRawData, m_PEHeader->OptionalHeader.SectionAlignment);
		Section->SizeOfRawData += Offset;
		m_PEHeader->OptionalHeader.SizeOfImage += ALIGN(Section->SizeOfRawData, m_PEHeader->OptionalHeader.SectionAlignment);
		//this should always be true...
		if (Section->Characteristics & IMAGE_SCN_CNT_INITIALIZED_DATA) m_PEHeader->OptionalHeader.SizeOfInitializedData += Offset;
		if (!ce_PTR && !SetBufferLen(Section->PointerToRawData + Section->SizeOfRawData)) return false;
	}

	m_RelocTableCapacity = BytesToAdd;
	vrprintf(2, "done!\n");
	return true;
}

/** Updates relocation table.
 * @return True on success, false otherwise.
 */
bool CPEFile::UpdateRelocTable()
{
	vrprintf(2, "Attempting to update relocation table... ");
	//sort relocs (reverse order)
	sort(m_NewRelocs.begin(), m_NewRelocs.end(), std::greater<DWORD>());

	DWORD RelocsRVA = m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress;

	char * raw = (char *)((DWORD)RvaToOffset(RelocsRVA) + m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size);
	DWORD start = (DWORD)raw;

	//find last 'RVA of block'
	DWORD lastblockRVA = 0;
	IMAGE_BASE_RELOCATION * Relocs = NULL;

	while (m_NewRelocs.size())
	{
		//check if allowed size is not exceeded
		if (((DWORD)raw - start) >= m_RelocTableCapacity)
		{
			vrprintf(2, "failed! Capacity reached.\n");
			vrprintf(3, "%u entries didn't fit\n", m_NewRelocs.size());
			return false;
		}
		//check if new block needed
		if (lastblockRVA != (m_NewRelocs.back() & 0xfffff000))
		{
			//pad to DWORD
			if (lastblockRVA)
			{
				if ((Relocs->SizeOfBlock - sizeof(IMAGE_BASE_RELOCATION)) % sizeof(DWORD))
				{
					*(WORD *)raw = 0;
					raw += sizeof(WORD);
				}
			}
			//create new block
			Relocs = (IMAGE_BASE_RELOCATION *) raw;
			Relocs->SizeOfBlock = sizeof(IMAGE_BASE_RELOCATION);
			lastblockRVA = Relocs->VirtualAddress = m_NewRelocs.back() & 0xfffff000;
			raw += sizeof(IMAGE_BASE_RELOCATION);
		}

		//add new entry
		*(WORD *)raw = (WORD)(m_NewRelocs.back() & 0x00000fff) | (IMAGE_REL_BASED_HIGHLOW << 12);
		raw += sizeof(WORD);
		Relocs->SizeOfBlock += sizeof(WORD);

		m_NewRelocs.pop_back();
	}

	//pad to DWORD
	if (lastblockRVA)
	{
		if ((Relocs->SizeOfBlock - sizeof(IMAGE_BASE_RELOCATION)) % sizeof(DWORD))
		{
			*(WORD *)raw = 0;
			raw += sizeof(WORD);
			Relocs->SizeOfBlock += sizeof(WORD);
		}
	}

	//update size data
	RvaToSection(RelocsRVA)->Misc.VirtualSize += (DWORD)raw - start;
	m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size += (DWORD)raw - start;

	vrprintf(2, "done!\n");
	return true;
}

/** Strips debug information. */
void CPEFile::RemoveDebug()
{
	DWORD Debug_Directory_RVA = m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].VirtualAddress;
	if (!Debug_Directory_RVA) return;

	IMAGE_DEBUG_DIRECTORY * DebugDir = (IMAGE_DEBUG_DIRECTORY *) RvaToOffset(Debug_Directory_RVA);

	if (DebugDir[0].PointerToRawData) SetBufferLen(DebugDir[0].PointerToRawData);

	m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].VirtualAddress = 0;
	m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size = 0;
}

/** Retrieves RVA of specified export name in exports table.
 * @param Function Function name.
 * @return RVA of specified export name in exports table.
 */
DWORD CPEFile::GetApiExportNameRVA(const char* Function)
{
	IMAGE_EXPORT_DIRECTORY *Exports = ( IMAGE_EXPORT_DIRECTORY * ) RvaToOffset( m_PEHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress );

	if ( Exports )
	{
		DWORD *Names = (DWORD *) RvaToOffset( Exports->AddressOfNames );

		for ( DWORD i = 0; i < Exports->NumberOfNames; i++ )
		{
			if ( ! strncmp( (char *) RvaToOffset( *Names), Function, strlen(Function) ) )
			{
				return *Names;
			}

			Names++;
		}

	}

	return 0;
}

