/*
 *  KernelEx
 *  Copyright (C) 2007, Xeno86
 *
 *  This file is part of KernelEx source code.
 *
 *  KernelEx is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation; version 2 of the License.
 *
 *  KernelEx is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/** @file wininit.h
 * Class WININIT.
 */

#ifndef __WININIT_H
#define __WININIT_H

#include <stack>

/** @class WININIT
 * Maintains wininit.ini entries.
 */
class WININIT
{
	/** @struct _2files
	 * Helper struct holding two filenames.
	 */
	struct _2files
	{
		char dest[MAX_PATH], src[MAX_PATH];
	};

	/** Stack holding wininit.ini entries. */
	std::stack<_2files> stk;
	
	/** Holds path to wininit.ini file. */
	char InitFile[MAX_PATH];

public:

	/** Constructor. */
	WININIT()
	{
		if (GetWindowsDirectory(InitFile, MAX_PATH) > 3) strcat(InitFile, "\\");
		strcat(InitFile, "WININIT.INI");
	}

	/** Destructor. */
	~WININIT()
	{
		clear();
	}

	/** Deletes all files not saved by save() method. @see save */
	void clear()
	{
		if (!stk.empty()) vrprintf(2, "\nRemoving temporary files ...\n");
		while (!stk.empty())
		{
			DeleteFile(stk.top().src);
			vrprintf(2, "... %s\n", stk.top().src);
			stk.pop();
		}
	}

	/** Checks if wininit.ini file exists.
	 * @return true if exists, false otherwise.
	 */
	bool exists()
	{
		return (GetFileAttributes(InitFile) != 0xffffffff);
	}

	/** Pushes new source-destination pair onto the stack.
	 * @param dest Destination file name (or 'NUL' for delete).
	 * @param src Source file name.
	 */
	void add(const char dest[], const char src[])
	{
		_2files t;
		strncpy(t.dest, dest, MAX_PATH);
		strncpy(t.src, src, MAX_PATH);
		stk.push(t);
	}

	/** Saves all data from stack to wininit.ini file. */
	void save()
	{
		FILE * fInit = NULL;
		if (!stk.empty())
		{
			fInit = fopen(InitFile, "w");
			fprintf(fInit, "[Rename]\n");
			vrprintf(2, "\nSaving WININIT.INI entries ...\n");
		}
		while (!stk.empty())
		{
			fprintf(fInit, "%s=%s\n", stk.top().dest, stk.top().src);
			vrprintf(2, "... %s => %s\n", stk.top().src, stk.top().dest);
			stk.pop();
		}
		if (fInit) fclose(fInit);
	}
} Wininit;

//we want to delete temporary files before closing the log... so we add this with atexit()
void __cdecl WininitClear(void)
{
	Wininit.clear();
}

#endif // __WININIT_H
